"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const SemanticVersion_1 = require("../Utils/SemanticVersion");
const TypeChecks_1 = require("../Utils/TypeChecks");
const Interfaces_1 = require("./Interfaces");
class DesignDocSync {
    constructor(provider, manifest) {
        this.provider = provider;
        this.manifest = manifest;
    }
    sync() {
        return __awaiter(this, void 0, void 0, function* () {
            for (const docName of Object.keys(this.manifest)) {
                yield this.syncDesignDoc(docName);
            }
        });
    }
    syncDesignDoc(docName) {
        return __awaiter(this, void 0, void 0, function* () {
            const docId = `_design/${docName}`;
            let designDocResult = null;
            try {
                designDocResult = yield this.provider.get(docId);
            }
            catch (e) { }
            const docType = this.manifest[docName].type;
            const createNewDoc = !Boolean(designDocResult);
            const designDoc = (createNewDoc
                ? DesignDocSync.NewDesignDoc(docType)
                : designDocResult);
            if (!designDoc.hasOwnProperty('views')) {
                designDoc.views = {};
            }
            const currentViews = designDoc.views;
            const newViews = this.getViews(docName);
            if (createNewDoc) {
                designDoc.views = newViews;
                if (DesignDocSync.IsMangoDesignDoc(designDoc)) {
                    DesignDocSync.EnsureMangoManifestOptionsObjects(designDoc.views);
                }
                yield this.provider.insert(designDoc, docId);
            }
            else if (DesignDocSync.ViewsWereUpdated(currentViews, newViews)) {
                const timeStamp = new Date().toJSON();
                designDoc.last_updated = timeStamp;
                designDoc.views = Object.assign({}, designDoc.views, DesignDocSync.MergeViews(currentViews, newViews));
                if (DesignDocSync.IsMangoDesignDoc(designDoc)) {
                    DesignDocSync.EnsureMangoManifestOptionsObjects(designDoc.views);
                }
                yield this.provider.update(designDoc, docId);
            }
            else if (TypeChecks_1.IsAnEmptyObject(designDoc.views)) {
                const timeStamp = new Date().toJSON();
                designDoc.last_updated = timeStamp;
                if (DesignDocSync.IsMangoDesignDoc(designDoc)) {
                    DesignDocSync.EnsureMangoManifestOptionsObjects(designDoc.views);
                }
                yield this.provider.update(designDoc, docId);
            }
        });
    }
    getViews(docName) {
        const viewManifestMap = this.manifest[docName].views;
        if (!viewManifestMap || typeof viewManifestMap !== 'object' || TypeChecks_1.IsAnEmptyObject(viewManifestMap)) {
            return {};
        }
        const viewNames = Object.keys(viewManifestMap);
        for (const viewName of viewNames) {
            DesignDocSync.ValidateView(docName, viewName, viewManifestMap[viewName]);
        }
        return viewManifestMap;
    }
    static NewDesignDoc(type) {
        const timeStamp = new Date().toJSON();
        const language = type === Interfaces_1.DesignDocType.JAVASCRIPT ? Interfaces_1.DesignDocLanguage.JAVASCRIPT : Interfaces_1.DesignDocLanguage.MANGO;
        return { date_created: timeStamp, last_updated: timeStamp, views: {}, language };
    }
    static MergeViews(oldViews, newViews) {
        const mergedViewMap = {};
        for (const viewName of Object.keys(newViews)) {
            if (!oldViews[viewName]) {
                mergedViewMap[viewName] = newViews[viewName];
            }
            else {
                mergedViewMap[viewName] = this.MergeView(oldViews[viewName], newViews[viewName]);
            }
        }
        return mergedViewMap;
    }
    static MergeView(oldView, newView) {
        if (!oldView.hasOwnProperty('version')) {
            oldView.version = '0.0.1';
        }
        const reuslt = SemanticVersion_1.VersionComparison(oldView.version, newView.version);
        switch (reuslt) {
            case SemanticVersion_1.VersionComparisonResult.LESS_THAN:
                return newView;
            case SemanticVersion_1.VersionComparisonResult.GREATER_THAN:
                return oldView;
            case SemanticVersion_1.VersionComparisonResult.EQUAL_TO:
                return Object.assign({}, newView, oldView);
        }
    }
    static ValidateView(docName, viewName, view) {
        if (!docName || !viewName || !view) {
            throw new Interfaces_1.DesignDocSyncException(`Invalid view. Either the docName, viewName, or view were invalid.
         docName: ${docName}    viewName: ${viewName}    view: ${view}`);
        }
        if (!view.version) {
            throw new Interfaces_1.DesignDocSyncException(`All Design Document Views must be versioned.
         A View named ${viewName} in the ${docName} design document did not include a version.`);
        }
        this.ValidateViewVersion(docName, viewName, view.version);
        if (this.IsJsViewManifest(view)) {
            this.ValidateJsView(docName, viewName, view);
        }
        else if (this.IsMangoViewManifest(view)) {
            this.ValidateMangoView(docName, viewName, view);
        }
        else {
            throw new Interfaces_1.DesignDocSyncException(`Invalid Design Doc View.
         A View named ${viewName} in the ${docName} design document contained the following invalid data: ${view}.`);
        }
    }
    static ValidateViewVersion(docName, viewName, version) {
        if (typeof version !== 'string') {
            throw new Interfaces_1.DesignDocSyncException(`All Design Document View versions must be a string value.
         A View named ${viewName} in the ${docName} design document supplied the following as a version: ${version}.`);
        }
        try {
            SemanticVersion_1.ValidateVersion(version);
        }
        catch (e) {
            throw new Interfaces_1.DesignDocSyncException(`All Design Document View versions must have a semantic version with a MAJOR.MINOR.PATCH format (e.g, 1.0.42).
         A View named ${viewName} in the ${docName} design document supplied the following as a version: ${version}.`);
        }
    }
    static ValidateJsView(docName, viewName, item) {
        if (!item.hasOwnProperty('map') ||
            ((TypeChecks_1.IsFalsy(item.map) && TypeChecks_1.IsNotAString(item.map)) || item.map.length === 0)) {
            throw new Interfaces_1.DesignDocSyncException(`All JavaScript Design Document View map functions must be a string value.
         A View named ${viewName} in the ${docName} design document supplied the following as a map function: ${item.map}.`);
        }
        if (item.hasOwnProperty('reduce') && (TypeChecks_1.IsNotAString(item.reduce) || TypeChecks_1.IsAnEmptyString(item.reduce))) {
            throw new Interfaces_1.DesignDocSyncException(`All JavaScript Design Document reduce fields must be a string.
         A View named ${viewName} in the ${docName} design document supplied the following as a reduce field: ${item.reduce}.`);
        }
    }
    static ValidateMangoView(docName, viewName, item) {
        if (!item.hasOwnProperty('map') || TypeChecks_1.IsFalsy(item.map)) {
            throw new Interfaces_1.DesignDocSyncException(`All Mango Design Document View must have a map object.
         A View named ${viewName} in the ${docName} design document supplied the following as a map object: ${item.map}.`);
        }
        if (!item.map.hasOwnProperty('fields') || TypeChecks_1.IsFalsy(item.map.fields)) {
            throw new Interfaces_1.DesignDocSyncException(`All Mango Design Document map objects must contain a field object.
         A View named ${viewName} in the ${docName} design document supplied the following as a map object: ${item.map}.`);
        }
        const fieldCount = Object.keys(item.map.fields).length;
        if (fieldCount < 1) {
            throw new Interfaces_1.DesignDocSyncException(`All Mango Design Document map objects must contain a field object with atleast one field.
      A View named ${viewName} in the ${docName} design document supplied the following as a map object: ${item.map}.`);
        }
        if (item.hasOwnProperty('reduce') && (TypeChecks_1.IsNotAString(item.reduce) || TypeChecks_1.IsAnEmptyString(item.reduce))) {
            throw new Interfaces_1.DesignDocSyncException(`All Mango Design Document reduce fields must be a string.
         A View named ${viewName} in the ${docName} design document supplied the following as a reduce field: ${item.reduce}.`);
        }
    }
    static ViewsWereUpdated(oldViews, newViews) {
        for (const viewName of Object.keys(oldViews)) {
            if (!newViews[viewName]) {
                return true;
            }
        }
        for (const viewName of Object.keys(newViews)) {
            if (!oldViews[viewName]) {
                return true;
            }
            if (this.ViewWasUpdated(viewName, oldViews[viewName], newViews[viewName])) {
                return true;
            }
        }
        return false;
    }
    static ViewWasUpdated(viewName, oldView, newView) {
        if (!oldView.version || newView.version > oldView.version) {
            return true;
        }
        if (oldView.version === newView.version) {
            if (this.IsJsViewManifest(oldView) && this.IsJsViewManifest(newView)) {
                if (oldView.map !== newView.map) {
                    return true;
                }
                else if (oldView.hasOwnProperty('reduce') &&
                    (!newView.hasOwnProperty('reduce') || oldView.reduce !== newView.reduce)) {
                    return true;
                }
                else if (newView.hasOwnProperty('reduce') &&
                    (!oldView.hasOwnProperty('reduce') || oldView.reduce !== newView.reduce)) {
                    return true;
                }
            }
            else if (this.IsMangoViewManifest(oldView) && this.IsMangoViewManifest(newView)) {
                if (oldView.hasOwnProperty('reduce') &&
                    (!newView.hasOwnProperty('reduce') || oldView.reduce !== newView.reduce)) {
                    return true;
                }
                else if (newView.hasOwnProperty('reduce') &&
                    (!oldView.hasOwnProperty('reduce') || oldView.reduce !== newView.reduce)) {
                    return true;
                }
                else if (this.MangoViewMapItemWasUpdated(oldView.map, newView.map)) {
                    return true;
                }
            }
            else {
                throw new Interfaces_1.DesignDocSyncException(`The old and new versions of the ${viewName} view are incompatible types.`);
            }
        }
        return false;
    }
    static MangoViewMapItemWasUpdated({ fields: oldFields }, { fields: newFields }) {
        for (const field of Object.keys(oldFields)) {
            if (!newFields.hasOwnProperty(field) || oldFields[field] !== newFields[field]) {
                return true;
            }
        }
        for (const field of Object.keys(newFields)) {
            if (!oldFields.hasOwnProperty(field) || newFields[field] !== oldFields[field]) {
                return true;
            }
        }
        return false;
    }
    static EnsureMangoManifestOptionsObjects(viewManifest) {
        for (const viewName of Object.keys(viewManifest)) {
            const view = viewManifest[viewName];
            this.EnsureMangoOptionsObject(view);
        }
    }
    static EnsureMangoOptionsObject(viewObject) {
        if (!viewObject.hasOwnProperty('options')) {
            this.BuildMangoOptionsObject(viewObject);
        }
        else {
            if (!this.IsValidMangoOptionsObject(viewObject)) {
                this.BuildMangoOptionsObject(viewObject);
            }
        }
    }
    static IsValidMangoOptionsObject(viewObject) {
        const oldOptions = viewObject.options;
        if (!oldOptions.hasOwnProperty('def') || TypeChecks_1.IsFalsy(oldOptions.def) || TypeChecks_1.IsAnEmptyObject(oldOptions.def)) {
            return false;
        }
        if (!oldOptions.def.hasOwnProperty('fields') ||
            TypeChecks_1.IsFalsy(oldOptions.def.fields) ||
            TypeChecks_1.IsAnEmptyObject(oldOptions.def.fields)) {
            return false;
        }
        const oldFields = oldOptions.def.fields;
        const currentFields = Object.keys(viewObject.map.fields);
        for (const field of oldFields) {
            if (!currentFields.includes(field)) {
                return false;
            }
        }
        for (const field of currentFields) {
            if (!oldFields.includes(field)) {
                return false;
            }
        }
        return true;
    }
    static BuildMangoOptionsObject(viewObject) {
        viewObject.options = { def: { fields: Object.keys(viewObject.map.fields).map(key => key) } };
    }
    static IsJsDesignDocManifest(designDoc) {
        if (designDoc.type !== Interfaces_1.DesignDocType.MANGO) {
            return false;
        }
        if (!designDoc.hasOwnProperty('views') || !this.IsJsViewManifestMap(designDoc.views)) {
            return false;
        }
        return true;
    }
    static IsJsViewManifestMap(manifestMap) {
        const keys = Object.keys(manifestMap);
        if (keys.length < 1) {
            return false;
        }
        for (const key of keys) {
            if (!this.IsJsViewManifest(manifestMap[key])) {
                return false;
            }
        }
        return true;
    }
    static IsJsViewManifest(manifest) {
        return manifest.hasOwnProperty('map') && this.IsJsViewMapItem(manifest.map);
    }
    static IsJsViewMapItem(item) {
        return TypeChecks_1.IsString(item);
    }
    static IsMangoDesignDoc(designDoc) {
        return designDoc.language === Interfaces_1.DesignDocLanguage.MANGO;
    }
    static IsMangoDesignDocManifest(designDoc) {
        if (designDoc.type !== Interfaces_1.DesignDocType.MANGO) {
            return false;
        }
        if (!designDoc.hasOwnProperty('views') || !this.IsMangoViewManifestMap(designDoc.views)) {
            return false;
        }
        return true;
    }
    static IsMangoViewManifestMap(manifestMap) {
        const keys = Object.keys(manifestMap);
        if (keys.length < 1) {
            return false;
        }
        for (const key of keys) {
            if (!this.IsMangoViewManifest(manifestMap[key])) {
                return false;
            }
        }
        return true;
    }
    static IsMangoViewManifest(manifest) {
        return manifest.hasOwnProperty('map') && this.IsMangoViewMapItem(manifest.map);
    }
    static IsMangoViewMapItem(item) {
        return TypeChecks_1.IsTruthy(item) && item.hasOwnProperty('fields');
    }
}
exports.DesignDocSync = DesignDocSync;
