"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const CouchCollation_1 = require("../Utils/CouchCollation");
const TypeChecks_1 = require("../Utils/TypeChecks");
const MockMango_1 = require("./MockMango");
class MockFind {
    constructor() { }
    static Find(query, docs) {
        const searchResults = MockMango_1.MockMango.Search(query.selector, docs);
        const sortedResults = SortAndFilterByFields(searchResults, query);
        const limitedResults = LimitAndSkip(sortedResults, query);
        const trimmedResults = FilterByFields(limitedResults, query);
        return trimmedResults;
    }
}
exports.MockFind = MockFind;
function LimitAndSkip(docs, { limit, skip }) {
    const docLengh = docs.length;
    let startIndex = 0;
    let endIndex = docLengh;
    if (skip) {
        startIndex += skip;
    }
    if (limit && endIndex - startIndex > limit) {
        endIndex = startIndex + limit;
    }
    if (startIndex === 0 && endIndex === docs.length) {
        return docs;
    }
    return docs.slice(startIndex, endIndex);
}
function FilterByFields(docs, { fields }) {
    if (!fields) {
        return docs;
    }
    if (!TypeChecks_1.IsArray(fields)) {
        throw new Error('The fields options for a search query must be an array of strings.');
    }
    if (fields.length === 0) {
        return docs;
    }
    const newDocs = [];
    const fieldFilterMap = GetFieldFilterMap(fields);
    const fieldFilterKeys = Object.keys(fieldFilterMap);
    for (let i = 0; i < docs.length; i++) {
        const newDoc = {};
        for (let j = 0; j < fieldFilterKeys.length; j++) {
            const fieldKey = fieldFilterKeys[j];
            const fieldKeyItem = fieldFilterMap[fieldKey];
            if (TypeChecks_1.IsDefined(docs[i][fieldKey])) {
                if (fieldKeyItem.length === 0) {
                    newDoc[fieldKey] = docs[i][fieldKey];
                    continue;
                }
                const lastIndex = fieldKeyItem.length - 1;
                let docScope = docs[i][fieldKey];
                for (let k = 0; k < fieldKeyItem.length; k++) {
                    if (TypeChecks_1.IsUndefined(docScope[fieldKeyItem[k]])) {
                        break;
                    }
                    if (k === lastIndex) {
                        newDoc[fieldKey] = { [fieldKeyItem[k]]: docScope[fieldKeyItem[k]] };
                    }
                    docScope = docScope[fieldKeyItem[k]];
                }
            }
        }
        newDocs[i] = newDoc;
    }
    return newDocs;
}
function GetFieldFilterMap(fields) {
    const fieldFilterMap = {};
    for (let i = 0; i < fields.length; i++) {
        const currentField = fields[i];
        if (TypeChecks_1.IsNotAString(currentField) || currentField === '') {
            throw new Error(`Invalid field option: ${currentField}. All field options must be a string.`);
        }
        const periodIndex = currentField.indexOf('.');
        if (periodIndex === -1 && !fieldFilterMap[currentField]) {
            fieldFilterMap[currentField] = [];
            continue;
        }
        const filterItems = currentField.split('.');
        if (filterItems.length < 1 || filterItems.includes('')) {
            throw new Error(`Invalid field detected in the fields option of the search query: ${currentField}`);
        }
        let currentFieldItem = fieldFilterMap[filterItems[0]];
        if (!currentFieldItem || currentFieldItem.length > filterItems.length - 1) {
            currentFieldItem = [];
            for (let j = 1; j < filterItems.length; j++) {
                currentFieldItem[j - 1] = filterItems[j];
            }
            fieldFilterMap[filterItems[0]] = currentFieldItem;
            continue;
        }
    }
    return fieldFilterMap;
}
function SortAndFilterByFields(docs, { sort }) {
    if (!sort) {
        return docs;
    }
    if (!TypeChecks_1.IsArray(sort)) {
        throw new Error(`Sort must be an array of sort specs, not ${sort}`);
    }
    if (sort.length === 0 || docs.length < 2) {
        return docs;
    }
    const docsLength = docs.length;
    const sortItemLength = sort.length;
    const filteredDocs = [];
    const docMap = {};
    let direction = 'asc';
    if (TypeChecks_1.IsObject(sort[0])) {
        const firstSortItemKeys = Object.keys(sort[0]);
        if (!firstSortItemKeys || firstSortItemKeys.length !== 1) {
            throw new Error(`Invalid sort field: ${sort[0]}`);
        }
        const sortField = firstSortItemKeys[0];
        if (firstSortItemKeys[sortField] !== 'asc' && firstSortItemKeys[sortField] !== 'desc') {
            throw new Error(`Invalid sort field: ${sort[0]}`);
        }
        direction = sort[0][sortField];
    }
    for (let i = 0; i < docsLength; i++) {
        const newItems = [];
        for (let j = 0; j < sortItemLength; j++) {
            const currentSortField = sort[j];
            let narrowedObject = false;
            if (TypeChecks_1.IsObject(currentSortField)) {
                const sortFieldKeys = Object.keys(currentSortField);
                if (!sortFieldKeys || sortFieldKeys.length !== 1) {
                    throw new Error(`Invalid sort field: ${currentSortField}`);
                }
                const sortField = sortFieldKeys[0];
                if (currentSortField[sortField] !== 'asc' && currentSortField[sortField] !== 'desc') {
                    throw new Error(`Invalid sort field: ${currentSortField}`);
                }
                if (currentSortField[sortField] !== direction) {
                    throw new Error('Sorts currently only support a single direction for all fields');
                }
                narrowedObject = NarrowObjectScope(sortField, docs[i]);
            }
            else if (TypeChecks_1.IsString(currentSortField) && currentSortField.length > 0) {
                if (direction === 'desc') {
                    throw new Error('Sorts currently only support a single direction for all fields');
                }
                narrowedObject = NarrowObjectScope(currentSortField, docs[i]);
            }
            else {
                throw new Error(`Invalid sort field: ${currentSortField}`);
            }
            if (narrowedObject) {
                newItems[j] = narrowedObject.doc[narrowedObject.field];
            }
            else {
                break;
            }
        }
        if (newItems.length === sortItemLength) {
            filteredDocs.push({ id: docs[i]._id, items: newItems });
            docMap[docs[i]._id] = docs[i];
        }
    }
    if (filteredDocs.length > 1) {
        if (direction === 'asc') {
            filteredDocs.sort((left, right) => {
                for (let i = 0; i < sortItemLength; i++) {
                    const comparison = CouchCollation_1.CompareValues(left.items[i], right.items[i]);
                    if (comparison !== 0) {
                        return comparison;
                    }
                }
                return 0;
            });
        }
        else {
            filteredDocs.sort((left, right) => {
                for (let i = 0; i < sortItemLength; i++) {
                    const comparison = -CouchCollation_1.CompareValues(left, right);
                    if (comparison !== 0) {
                        return comparison;
                    }
                }
                return 0;
            });
        }
    }
    const sortedDocs = [];
    for (let i = 0; i < filteredDocs.length; i++) {
        sortedDocs[i] = docMap[filteredDocs[i].id];
    }
    return sortedDocs;
}
function NarrowObjectScope(field, doc) {
    const periodIndex = field.indexOf('.');
    if (periodIndex === -1) {
        if (doc.hasOwnProperty(field)) {
            return { field, doc };
        }
        return false;
    }
    const nestedFields = field.split('.');
    if (nestedFields.length < 1 || nestedFields.includes('')) {
        throw new Error(`Invalid nested field detected: ${field}`);
    }
    if (!doc.hasOwnProperty(nestedFields[0])) {
        return false;
    }
    let currentObjectScope = doc;
    let currentField = nestedFields[0];
    let nestedObjectScope = doc[nestedFields[0]];
    for (let j = 1; j < nestedFields.length; j++) {
        if (!nestedObjectScope.hasOwnProperty(nestedFields[j])) {
            return false;
        }
        currentObjectScope = nestedObjectScope;
        nestedObjectScope = currentObjectScope[nestedFields[j]];
        currentField = nestedFields[j];
    }
    return { field: currentField, doc: currentObjectScope };
}
