"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const CouchCollation_1 = require("../Utils/CouchCollation");
const RegexTranslator_1 = require("../Utils/RegexTranslator");
const TypeChecks_1 = require("../Utils/TypeChecks");
const ALL_OPERATORS = [
    '$and',
    '$or',
    '$not',
    '$nor',
    '$all',
    '$elemMatch',
    '$allMatch',
    '$lt',
    '$lte',
    '$eq',
    '$ne',
    '$gte',
    '$gt',
    '$exists',
    '$type',
    '$in',
    '$nin',
    '$size',
    '$mod',
    '$regex',
];
const CONDITION_OPERATORS = [
    '$lt',
    '$lte',
    '$eq',
    '$ne',
    '$gte',
    '$gt',
    '$exists',
    '$type',
    '$in',
    '$nin',
    '$size',
    '$mod',
    '$regex',
];
class MockMango {
    constructor() { }
    static Search(selector, docs) {
        return FilterWithSelector(selector, docs);
    }
}
exports.MockMango = MockMango;
function FilterWithSelector(selector, docs) {
    const filtered = [];
    let filteredIndex = 0;
    const documentLength = docs.length;
    for (let i = 0; i < documentLength; i++) {
        if (ObjectMatchesSelector(selector, docs[i])) {
            filtered[filteredIndex] = docs[i];
            filteredIndex++;
        }
    }
    return filtered;
}
function ObjectMatchesSelector(selector, doc) {
    const operators = Object.keys(selector);
    const operatorLength = operators.length;
    for (let i = 0; i < operatorLength; i++) {
        if (CONDITION_OPERATORS.indexOf(operators[i]) !== -1) {
            throw new Error(`The condition operator ${operators[i]} was detected
      at the root scope of a Mango selector. This is not allowed.`);
        }
        const operator = operators[i];
        const operatorValue = selector[operator];
        if (!EvaluateOperator(null, operator, operatorValue, doc)) {
            return false;
        }
    }
    return true;
}
function NestedObjectMatchesSelector(field, selector, objectScope) {
    const operators = Object.keys(selector);
    const operatorLength = operators.length;
    for (let i = 0; i < operatorLength; i++) {
        const operator = operators[i];
        const operatorValue = selector[operator];
        if (!EvaluateOperator(field, operator, operatorValue, objectScope)) {
            return false;
        }
    }
    return true;
}
function EvaluateOperator(field, operator, operatorValue, objectScope) {
    if (operator.charAt(0) !== '$') {
        if (operator.indexOf('.') !== -1) {
            const newItems = ExpandFieldOperator(operator, objectScope);
            if (!newItems) {
                return false;
            }
            return EvaluateFieldOperator(newItems.newFieldOperator, operatorValue, newItems.newScope);
        }
        else {
            return EvaluateFieldOperator(operator, operatorValue, objectScope);
        }
    }
    return EvaluateMangoOperator(field, operator, operatorValue, objectScope);
}
function EvaluateFieldOperator(operator, operatorValue, objectScope) {
    if (!objectScope.hasOwnProperty(operator)) {
        return false;
    }
    if (!TypeChecks_1.IsObject(operatorValue)) {
        const comparisonResult = CouchCollation_1.CompareValues(objectScope[operator], operatorValue);
        return comparisonResult === 0;
    }
    const operatorValueKeys = Object.keys(operatorValue);
    const operatorValueKeyLength = operatorValueKeys.length;
    for (let i = 0; i < operatorValueKeyLength; i++) {
        if (ALL_OPERATORS.indexOf(operatorValueKeys[i]) !== -1) {
            return NestedObjectMatchesSelector(operator, operatorValue, objectScope);
        }
        else {
            return EvaluateFieldOperator(operatorValueKeys[i], operatorValue[operatorValueKeys[i]], objectScope[operator]);
        }
    }
    const objectComparisonResult = CouchCollation_1.CompareValues(objectScope[operator], operatorValue);
    return objectComparisonResult === 0;
}
function ExpandFieldOperator(operator, objectScope) {
    const expandedField = operator.split('.');
    const fieldLength = expandedField.length;
    let newFieldOperator = null;
    let newScope = objectScope;
    for (let i = 0; i < fieldLength; i++) {
        newFieldOperator = expandedField[i];
        if (!newScope.hasOwnProperty(newFieldOperator)) {
            return false;
        }
        if (fieldLength > 1 && i < fieldLength - 1) {
            newScope = newScope[newFieldOperator];
        }
    }
    return { newScope, newFieldOperator };
}
function EvaluateMangoOperator(field, operator, operatorValue, objectScope) {
    switch (operator) {
        case '$and':
            return EvaluateAndCombinationOperator(field, operatorValue, objectScope);
        case '$or':
            return EvaluateOrCombinationOperator(field, operatorValue, objectScope);
        case '$not':
            throw new Error('The $not operator has not yet been implemented');
        case '$nor':
            throw new Error('The $nor operator has not yet been implemented');
        case '$all':
            throw new Error('The $all operator has not yet been implemented');
        case '$elemMatch':
            return EvaluateElemMatchOperator(field, operatorValue, objectScope);
        case '$allMatch':
            throw new Error('The $allMatch operator has not yet been implemented');
        case '$lt':
            if (field) {
                return objectScope.hasOwnProperty(field) && CouchCollation_1.CompareValues(objectScope[field], operatorValue) === -1;
            }
            return CouchCollation_1.CompareValues(objectScope, operatorValue) === -1;
        case '$lte':
            if (field) {
                return objectScope.hasOwnProperty(field) && CouchCollation_1.CompareValues(objectScope[field], operatorValue) <= 0;
            }
            return CouchCollation_1.CompareValues(objectScope, operatorValue) <= 0;
        case '$eq':
            if (field) {
                return objectScope.hasOwnProperty(field) && CouchCollation_1.CompareValues(objectScope[field], operatorValue) === 0;
            }
            return CouchCollation_1.CompareValues(objectScope, operatorValue) === 0;
        case '$ne':
            if (field) {
                return objectScope.hasOwnProperty(field) && CouchCollation_1.CompareValues(objectScope[field], operatorValue) !== 0;
            }
            return CouchCollation_1.CompareValues(objectScope, operatorValue) !== 1;
        case '$gte':
            if (field) {
                return objectScope.hasOwnProperty(field) && CouchCollation_1.CompareValues(objectScope[field], operatorValue) >= 0;
            }
            return CouchCollation_1.CompareValues(objectScope, operatorValue) >= 0;
        case '$gt':
            if (field) {
                return objectScope.hasOwnProperty(field) && CouchCollation_1.CompareValues(objectScope[field], operatorValue) === 1;
            }
            return CouchCollation_1.CompareValues(objectScope, operatorValue) === 1;
        case '$exists':
            if (field) {
                return objectScope.hasOwnProperty(field) === operatorValue;
            }
            return TypeChecks_1.IsDefined(objectScope) === operatorValue;
        case '$type':
            throw new Error('The $type operator has not yet been implemented');
        case '$in':
            return EvalulateInOperator(field, operatorValue, objectScope);
        case '$nin':
            return EvalulateNinOperator(field, operatorValue, objectScope);
        case '$size':
            if (!TypeChecks_1.IsAPositiveInteger(operatorValue)) {
                throw new Error(`Bad argument for operator $size: ${operatorValue}`);
            }
            if (field) {
                if (objectScope.hasOwnProperty(field) ||
                    !TypeChecks_1.IsArray(objectScope[field]) ||
                    objectScope[field].length !== operatorValue) {
                    return false;
                }
                return true;
            }
            if (!TypeChecks_1.IsArray(objectScope) || objectScope.length !== operatorValue) {
                return false;
            }
            return true;
        case '$mod':
            throw new Error('The $mod operator has not yet been implemented');
        case '$regex':
            if (!TypeChecks_1.IsString(operatorValue)) {
                throw new Error('The value provided for the $regex operator must be a string');
            }
            const { pattern, flags } = RegexTranslator_1.ConvertInlineModifiers(operatorValue);
            if (field) {
                return (objectScope.hasOwnProperty(field) &&
                    TypeChecks_1.IsString(objectScope[field]) &&
                    objectScope[field].match(new RegExp(pattern, flags)));
            }
            return TypeChecks_1.IsString(objectScope) && objectScope.match(new RegExp(pattern, flags));
    }
    throw new Error(`${operator} is not a valid Mango operator!`);
}
function EvaluateAndCombinationOperator(field, operatorValue, objectScope) {
    if (!TypeChecks_1.IsArray(operatorValue)) {
        throw new Error('The operator value for the $and operator must be an array');
    }
    const andSelectorsLength = operatorValue.length;
    for (let i = 0; i < andSelectorsLength; i++) {
        if (!NestedObjectMatchesSelector(field, operatorValue[i], objectScope)) {
            return false;
        }
    }
    return true;
}
function EvaluateElemMatchOperator(field, operatorValue, objectScope) {
    if (!field) {
        throw new Error('The $elemMatch operator requires a field to be specified');
    }
    if (!TypeChecks_1.IsObject(operatorValue)) {
        throw new Error('The $elemMatch operator value must be a selector object');
    }
    if (!objectScope.hasOwnProperty(field) || !TypeChecks_1.IsArray(objectScope[field])) {
        return false;
    }
    if (TypeChecks_1.IsAnEmptyObject(operatorValue)) {
        return false;
    }
    const arrayItemsLength = objectScope[field].length;
    for (let i = 0; i < arrayItemsLength; i++) {
        if (NestedObjectMatchesSelector(null, operatorValue, objectScope[field][i])) {
            return true;
        }
    }
    return false;
}
function EvalulateInOperator(field, operatorValue, objectScope) {
    if (!TypeChecks_1.IsArray(operatorValue)) {
        throw new Error('The $in operator must contain a value that is an array');
    }
    if (field) {
        if (!objectScope.hasOwnProperty(field)) {
            return false;
        }
        if (TypeChecks_1.IsArray(objectScope[field])) {
            for (let i = 0; i < operatorValue.length; i++) {
                for (let j = 0; j < objectScope[field].length; j++) {
                    if (CouchCollation_1.CompareValues(objectScope[field][j], operatorValue[i]) === 0) {
                        return true;
                    }
                }
            }
        }
        else {
            for (let i = 0; i < operatorValue.length; i++) {
                if (CouchCollation_1.CompareValues(objectScope[field], operatorValue[i]) === 0) {
                    return true;
                }
            }
        }
        return false;
    }
    if (TypeChecks_1.IsArray(objectScope)) {
        for (let i = 0; i < operatorValue.length; i++) {
            for (let j = 0; j < objectScope.length; j++) {
                if (CouchCollation_1.CompareValues(objectScope[j], operatorValue[i]) === 0) {
                    return true;
                }
            }
        }
    }
    else {
        for (let i = 0; i < operatorValue.length; i++) {
            if (CouchCollation_1.CompareValues(objectScope, operatorValue[i]) === 0) {
                return true;
            }
        }
    }
    return false;
}
function EvalulateNinOperator(field, operatorValue, objectScope) {
    if (!TypeChecks_1.IsArray(operatorValue)) {
        throw new Error('The $nin operator must contain a value that is an array');
    }
    if (field) {
        if (!objectScope.hasOwnProperty(field)) {
            return false;
        }
        if (TypeChecks_1.IsArray(objectScope[field])) {
            for (let i = 0; i < operatorValue.length; i++) {
                for (let j = 0; j < objectScope[field].length; j++) {
                    if (CouchCollation_1.CompareValues(objectScope[field][j], operatorValue[i]) === 0) {
                        return false;
                    }
                }
            }
        }
        else {
            for (let i = 0; i < operatorValue.length; i++) {
                if (CouchCollation_1.CompareValues(objectScope[field], operatorValue[i]) === 0) {
                    return false;
                }
            }
        }
        return true;
    }
    if (TypeChecks_1.IsArray(objectScope)) {
        for (let i = 0; i < operatorValue.length; i++) {
            for (let j = 0; j < objectScope.length; j++) {
                if (CouchCollation_1.CompareValues(objectScope[j], operatorValue[i]) === 0) {
                    return false;
                }
            }
        }
    }
    else {
        for (let i = 0; i < operatorValue.length; i++) {
            if (CouchCollation_1.CompareValues(objectScope, operatorValue[i]) === 0) {
                return false;
            }
        }
    }
    return true;
}
function EvaluateOrCombinationOperator(field, operatorValue, objectScope) {
    if (!TypeChecks_1.IsArray(operatorValue)) {
        throw new Error('The operator value for the $or operator must be an array');
    }
    const andSelectorsLength = operatorValue.length;
    for (let i = 0; i < andSelectorsLength; i++) {
        if (NestedObjectMatchesSelector(field, operatorValue[i], objectScope)) {
            return true;
        }
    }
    return false;
}
function getUtilities() {
    return {
        FilterWithSelector,
    };
}
exports.getUtilities = getUtilities;
