"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const uuid = require("uuid");
const CouchTypes_1 = require("../Interfaces/CouchTypes");
const CouchCollation_1 = require("../Utils/CouchCollation");
const MockDesign_1 = require("./MockDesign");
const MockFind_1 = require("./MockFind");
const MockBulkTool_1 = require("./MockBulkTool");
function GenerateMockProvider(manifestLocator) {
    const factory = {
        create(databaseName) {
            try {
                const newProvider = new MockProvider(manifestLocator(databaseName));
                return Promise.resolve(newProvider);
            }
            catch (e) {
                throw new Error(`Failed to create a mock provider for the ${databaseName} CouchDB Database.
        ${e}`);
            }
        },
    };
    return factory;
}
exports.GenerateMockProvider = GenerateMockProvider;
class MockProviderManager {
    static AddInstance(name, instance) {
        const existingInstance = this.Instances.get(name);
        if (existingInstance) {
            throw new Error(`A Mock ${name} Couch Provider instance already exists.`);
        }
        this.Instances.set(name, instance);
    }
    static PreloadInstance(name, docs) {
        return __awaiter(this, void 0, void 0, function* () {
            const existingInstance = this.Instances.get(name);
            if (!existingInstance) {
                throw new Error(`Attempted to preload a Mock ${name} Couch Provider instance that does not exist.`);
            }
            for (const doc of docs) {
                if (doc.id) {
                    yield existingInstance.insert(doc.data, doc.id);
                }
                else {
                    yield existingInstance.insert(doc.data);
                }
            }
        });
    }
}
MockProviderManager.Instances = new Map();
exports.MockProviderManager = MockProviderManager;
class MockProvider {
    constructor(manifest) {
        this.designDocs = manifest ? MockDesign_1.MockDesign.Generate(manifest) : new Map();
        this.store = new Map();
        this.bulkTool = new MockBulkTool_1.MockBulkToolProvider(this);
    }
    getBulkTool() {
        return this.bulkTool;
    }
    get(id) {
        const doc = this.store.get(id);
        if (!doc) {
            throw new Error(CouchTypes_1.NOT_FOUND);
        }
        return Promise.resolve(JSON.parse(JSON.stringify(doc)));
    }
    insert(data, id) {
        if (!id) {
            id = uuid.v4();
        }
        const rev = uuid.v4();
        const newDocument = JSON.parse(JSON.stringify(data));
        newDocument._id = id;
        newDocument._rev = rev;
        this.store.set(id, newDocument);
        for (const designDoc of this.designDocs.values()) {
            designDoc.insert(newDocument);
        }
        return Promise.resolve({ ok: true, id, rev });
    }
    update(data, id, rev) {
        const doc = this.store.get(id);
        if (!doc) {
            throw new Error(CouchTypes_1.NOT_FOUND);
        }
        if (doc._rev !== rev) {
            throw new Error(CouchTypes_1.DOCUMENT_UPDATE_CONFLICT);
        }
        const updatedRev = uuid.v4();
        const updatedDocument = JSON.parse(JSON.stringify(data));
        updatedDocument._id = id;
        updatedDocument._rev = updatedRev;
        this.store.set(id, updatedDocument);
        for (const designDoc of this.designDocs.values()) {
            designDoc.update(updatedDocument);
        }
        return Promise.resolve({ ok: true, id, rev: updatedRev });
    }
    delete(id, rev) {
        const doc = this.store.get(id);
        if (!doc) {
            throw new Error(CouchTypes_1.NOT_FOUND);
        }
        if (doc._rev !== rev) {
            throw new Error(CouchTypes_1.DOCUMENT_UPDATE_CONFLICT);
        }
        const deletedRev = uuid.v4();
        this.store.delete(id);
        for (const designDoc of this.designDocs.values()) {
            designDoc.delete(id);
        }
        return Promise.resolve({ ok: true, id, rev: deletedRev });
    }
    view(designDocument, view, options) {
        const designDoc = this.designDocs.get(designDocument);
        if (!designDoc) {
            throw new Error(`No design document named: ${designDocument} exists!`);
        }
        const viewResults = designDoc.view(view, options);
        if (options && options.include_docs) {
            for (let i = 0; i < viewResults.rows.length; i++) {
                const viewResult = viewResults.rows[i];
                viewResult.doc = JSON.parse(JSON.stringify(this.store.get(viewResult.id)));
            }
        }
        return Promise.resolve(viewResults);
    }
    find(query) {
        const bookmark = uuid.v4();
        const docs = Array.from(this.store.values(), value => JSON.parse(JSON.stringify(value)));
        docs.sort((left, right) => CouchCollation_1.CompareValues(left._id, right._id));
        return Promise.resolve({ bookmark, docs: MockFind_1.MockFind.Find(query, docs) });
    }
    bulkUpdate(docs) {
        return true;
    }
    viewEngine(designDocument, aggregateViews) {
        return true;
    }
}
exports.MockProvider = MockProvider;
