"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const CouchCollation_1 = require("../Utils/CouchCollation");
const TypeChecks_1 = require("../Utils/TypeChecks");
const BuiltInReducers_1 = require("./BuiltInReducers");
class MockView {
    constructor(manifest) {
        this.currentItem = '-1';
        this.viewIndex = new Map();
        this.mapFunction = MapFunctionFromString(this, manifest.map, this.emit);
        if (manifest.reduce) {
            this.reduceFunction = ReduceFunctionFromString(manifest.reduce);
        }
    }
    get(id) {
        return this.viewIndex.get(id);
    }
    update(doc) {
        const wasDeleted = this.viewIndex.delete(doc._id);
        if (wasDeleted) {
            this.map(doc);
        }
    }
    delete(id) {
        this.viewIndex.delete(id);
    }
    map(doc) {
        this.currentItem = doc._id;
        this.mapFunction(doc);
        this.currentItem = '-1';
    }
    query(options) {
        if (!options) {
            if (this.reduceFunction) {
                const rows = this.reduce(this.collectIndexItems(), { group: false, rereduce: false });
                return { rows };
            }
            else {
                const rows = [...this.collectIndexItems()];
                const total_rows = rows.length;
                const offset = 0;
                return { total_rows, offset, rows };
            }
        }
        return this.queryWithOptions(options);
    }
    reduce(indexItems, { group, groupLevel, rereduce, sorted }) {
        if (!this.reduceFunction) {
            throw new Error('Attempted to use a reduce function on a view without one');
        }
        const indexItemCount = indexItems.length;
        if (!group) {
            const keys = indexItems.map(({ key }) => key);
            const values = indexItems.map(({ value }) => value);
            return [{ key: null, value: this.reduceFunction(keys, values, rereduce) }];
        }
        if (!sorted) {
            CouchCollation_1.SortViewItems(indexItems);
        }
        const groups = [];
        let currGroup = [];
        let prevItem = indexItems[0];
        let currItem = indexItems[0];
        currGroup.push(currItem);
        groups.push(currGroup);
        for (let i = 1; i < indexItemCount; i++) {
            currItem = indexItems[i];
            const matchesPrev = CouchCollation_1.CompareValues(prevItem.key, currItem.key, groupLevel);
            if (matchesPrev !== 0) {
                const newGroup = [];
                groups.push(newGroup);
                currGroup = newGroup;
            }
            currGroup.push(currItem);
            prevItem = indexItems[i];
        }
        const results = [];
        const groupsLength = groups.length;
        for (let i = 0; i < groupsLength; i++) {
            const keys = groups[i].map(({ key }) => key);
            const values = groups[i].map(({ value }) => value);
            let resultKey = keys[0];
            if (groupLevel) {
                if (TypeChecks_1.IsArray(resultKey)) {
                    resultKey = resultKey.slice(0, groupLevel);
                }
                else if (TypeChecks_1.IsObject(resultKey)) {
                    const keyFields = Object.keys(resultKey);
                    const filteredKey = {};
                    for (let j = 0; j < groupLevel; j++) {
                        filteredKey[keyFields[j]] = resultKey[keyFields[j]];
                    }
                    resultKey = filteredKey;
                }
            }
            results.push({ key: resultKey, value: this.reduceFunction(keys, values, rereduce) });
        }
        return results;
    }
    collectIndexItems() {
        const internalIndexItems = [];
        for (const [id, indexItems] of this.viewIndex) {
            for (const { key, value } of indexItems) {
                internalIndexItems.push({ id, key, value });
            }
        }
        return internalIndexItems;
    }
    emit(self) {
        return (key, value) => {
            if (self.currentItem === '-1') {
                throw new Error('Attempted to emit a value to an invalid index item');
            }
            const indexSection = self.viewIndex.get(self.currentItem);
            if (!indexSection) {
                const newIndexSection = [{ key, value }];
                self.viewIndex.set(self.currentItem, newIndexSection);
            }
            else {
                indexSection.push({ key, value });
            }
        };
    }
    queryWithOptions(options) {
        const initialRows = [...this.collectIndexItems()];
        const initialLength = initialRows.length;
        CouchCollation_1.SortViewItems(initialRows);
        if (options.descending) {
            initialRows.reverse();
        }
        const filterResults = this.filterIndexRowsWithKeyOptions(initialRows, options);
        if (options.reduce) {
            if (options.include_docs) {
                throw new Error('Reduced view queries are not allowed to include docs');
            }
            filterResults.rows = this.reduce(filterResults.rows, {
                sorted: true,
                group: options.group,
                groupLevel: options.group_level,
            });
        }
        const filteredSize = filterResults.rows.length;
        const startIndex = options.skip ? options.skip : 0;
        const skippedSize = filteredSize - startIndex;
        const offset = filterResults.offset + startIndex;
        let rows = null;
        if (options.limit && skippedSize > options.limit) {
            rows = filterResults.rows.slice(startIndex, options.limit + startIndex);
        }
        else {
            if (startIndex !== 0) {
                rows = filterResults.rows.slice(startIndex);
            }
            else {
                rows = filterResults.rows;
            }
        }
        if (!rows) {
            throw new Error(`An unexpected error occurred while trying to find rows.`);
        }
        if (options.reduce) {
            return { rows };
        }
        return { total_rows: initialLength, offset, rows };
    }
    filterIndexRowsWithKeyOptions(rows, options) {
        const itemLength = rows.length;
        if (options.keys) {
            if (options.key || options.startkey || options.endkey || options.inclusive_end) {
                throw new Error('`keys` is incompatible with `inclusive_end`, `key`, `start_key`, and `end_key`');
            }
            const filteredRows = [];
            const possibleKeysLength = options.keys.length;
            let filteredRowsIndex = 0;
            let offset = 0;
            for (let i = 0; i < itemLength; i++) {
                for (let j = 0; j < possibleKeysLength; j++) {
                    if (CouchCollation_1.CompareValues(rows[i].key, options.keys[j]) === 0) {
                        filteredRows[filteredRowsIndex] = rows[i];
                        if (filteredRowsIndex === 0) {
                            offset = i;
                        }
                        filteredRowsIndex++;
                        break;
                    }
                }
            }
            return { offset, rows: filteredRows };
        }
        let inclusiveEnd = true;
        let startKey = 0;
        let endKey = itemLength;
        if (typeof options.inclusive_end !== 'undefined') {
            inclusiveEnd = options.inclusive_end;
        }
        if (options.key) {
            let foundKey = false;
            for (let i = 0; i < itemLength; i++) {
                const comparison = CouchCollation_1.CompareValues(rows[i].key, options.key);
                if (comparison === 0) {
                    if (!foundKey) {
                        startKey = i;
                        foundKey = true;
                    }
                }
                else if (foundKey) {
                    endKey = i;
                    break;
                }
                if (i === itemLength - 1) {
                    startKey = -1;
                }
            }
        }
        else {
            if (options.startkey) {
                let maxDepth;
                if (TypeChecks_1.IsArray(options.startkey)) {
                    maxDepth = options.startkey.length;
                }
                else if (TypeChecks_1.IsObject(options.startkey)) {
                    maxDepth = Object.keys(options.startkey).length;
                }
                for (let i = 0; i < itemLength; i++) {
                    const comparison = CouchCollation_1.CompareValues(rows[i].key, options.startkey, maxDepth);
                    if (comparison !== -1) {
                        startKey = i;
                        break;
                    }
                    if (i === itemLength - 1) {
                        startKey = -1;
                    }
                }
            }
            if (options.endkey && startKey !== -1) {
                if (!inclusiveEnd) {
                    for (let i = 0; i < itemLength; i++) {
                        const comparison = CouchCollation_1.CompareValues(rows[i].key, options.endkey);
                        if (comparison === 0) {
                            endKey = i;
                            break;
                        }
                        else if (comparison === 1) {
                            endKey = i;
                            break;
                        }
                    }
                }
                else {
                    for (let i = 0; i < itemLength; i++) {
                        const comparison = CouchCollation_1.CompareValues(rows[i].key, options.endkey);
                        if (comparison === 1) {
                            endKey = i;
                            break;
                        }
                    }
                }
            }
            else if (!options.endkey && !inclusiveEnd) {
                throw new Error('`inclusive_end` is not allowed to be false without an `endkey`');
            }
        }
        if (startKey === -1) {
            return { offset: itemLength, rows: [] };
        }
        else if (startKey === 0 && endKey === itemLength) {
            return { offset: 0, rows };
        }
        if (endKey === itemLength) {
            return { offset: startKey, rows: rows.slice(startKey) };
        }
        return { offset: startKey, rows: rows.slice(startKey, endKey) };
    }
}
exports.MockView = MockView;
function FunctionFromString(str) {
    const sanitizedString = str.replace(/\r\n?|\n|\t/g, ' ');
    const matches = RegExp(/function *\(([^()]*)\)[ \n\t]*{(.*)}/gim).exec(sanitizedString);
    if (!matches)
        throw new Error(`${sanitizedString} is not a valid function!`);
    const func = new Function(matches[1].split(','), matches[2]);
    if (!func)
        throw new Error(`${sanitizedString} is not a valid function!`);
    return func;
}
function MapFunctionFromString(parent, str, emit) {
    const modifiedString = str.replace(/emit/g, 'this.emit');
    const func = FunctionFromString(modifiedString);
    const boundFunc = func.bind({ emit: emit(parent) });
    return boundFunc;
}
function ReduceFunctionFromString(str) {
    switch (str) {
        case '_count':
            return BuiltInReducers_1.Count;
        case '_stats':
            return BuiltInReducers_1.Stats;
        case '_sum':
            return BuiltInReducers_1.Sum;
        default:
            return FunctionFromString(str);
    }
}
function getUtilities() {
    return {
        MapFunctionFromString,
        ReduceFunctionFromString,
    };
}
exports.getUtilities = getUtilities;
