"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const TypeChecks_1 = require("./TypeChecks");
var KeyType;
(function (KeyType) {
    KeyType[KeyType["NULL"] = 1] = "NULL";
    KeyType[KeyType["BOOLEAN"] = 2] = "BOOLEAN";
    KeyType[KeyType["NUMBER"] = 3] = "NUMBER";
    KeyType[KeyType["STRING"] = 4] = "STRING";
    KeyType[KeyType["ARRAY"] = 5] = "ARRAY";
    KeyType[KeyType["OBJECT"] = 6] = "OBJECT";
})(KeyType || (KeyType = {}));
function SortViewItems(indexItems, direction = 'asc') {
    if (direction === 'asc') {
        indexItems.sort((left, right) => CompareValues(left.key, right.key));
    }
    else {
        indexItems.sort((left, right) => -CompareValues(left.key, right.key));
    }
}
exports.SortViewItems = SortViewItems;
function CompareValues(left, right, maxDepth) {
    const leftType = GetKeyType(left);
    const rightType = GetKeyType(right);
    if (leftType < rightType) {
        return -1;
    }
    else if (leftType > rightType) {
        return 1;
    }
    return CompareSameType(leftType, left, right, maxDepth);
}
exports.CompareValues = CompareValues;
function CompareSameType(type, left, right, maxDepth) {
    switch (type) {
        case KeyType.STRING:
            return CompareStrings(left, right);
        case KeyType.ARRAY:
            return CompareArrays(left, right, maxDepth);
        case KeyType.NUMBER:
            return CompareNumbers(left, right);
        case KeyType.OBJECT:
            return CompareObjects(left, right, maxDepth);
        case KeyType.NULL:
            return 0;
        case KeyType.BOOLEAN:
            return CompareBooleans(left, right);
    }
    throw new Error(`${type} is not a valid View Index Item key type`);
}
function CompareStrings(left, right) {
    return left.localeCompare(right);
}
function CompareArrays(left, right, maxDepth) {
    const leftSize = left.length;
    const rightSize = right.length;
    const leftLimit = maxDepth ? (maxDepth > leftSize ? leftSize : maxDepth) : leftSize;
    for (let i = 0; i < leftLimit; i++) {
        if (!right[i]) {
            return 1;
        }
        const leftPrecedence = CompareValues(left[i], right[i]);
        if (leftPrecedence === -1) {
            return -1;
        }
        else if (leftPrecedence === 1) {
            return 1;
        }
    }
    if (!maxDepth && rightSize > leftSize) {
        return -1;
    }
    return 0;
}
function CompareNumbers(left, right) {
    if (left === right) {
        return 0;
    }
    return left > right ? 1 : -1;
}
function CompareObjects(left, right, maxDepth) {
    const leftKeys = Object.keys(left);
    const leftSize = leftKeys.length;
    const rightKeys = Object.keys(right);
    const rightSize = rightKeys.length;
    const leftLimit = maxDepth ? (maxDepth > leftSize ? leftSize : maxDepth) : leftSize;
    for (let i = 0; i < leftLimit; i++) {
        if (!rightKeys[i]) {
            return 1;
        }
        const leftKeyPrecedence = CompareStrings(leftKeys[i], rightKeys[i]);
        if (leftKeyPrecedence === -1) {
            return -1;
        }
        else if (leftKeyPrecedence === 1) {
            return 1;
        }
        const leftValuePrecedence = CompareValues(left[leftKeys[i]], right[rightKeys[i]]);
        if (leftValuePrecedence === -1) {
            return -1;
        }
        else if (leftValuePrecedence === 1) {
            return 1;
        }
    }
    if (!maxDepth && rightSize > leftSize) {
        return -1;
    }
    return 0;
}
function CompareBooleans(left, right) {
    if (left === right) {
        return 0;
    }
    return left > right ? 1 : -1;
}
function GetKeyType(key) {
    if (TypeChecks_1.IsString(key)) {
        return KeyType.STRING;
    }
    else if (TypeChecks_1.IsArray(key)) {
        return KeyType.ARRAY;
    }
    else if (TypeChecks_1.IsNumber(key)) {
        return KeyType.NUMBER;
    }
    else if (TypeChecks_1.IsObject(key)) {
        return KeyType.OBJECT;
    }
    else if (TypeChecks_1.IsNull(key)) {
        return KeyType.NULL;
    }
    else if (TypeChecks_1.IsBoolean(key)) {
        return KeyType.BOOLEAN;
    }
    throw new Error(`Invalid View Index Item Key: ${key}`);
}
