"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const TypeChecks_1 = require("./TypeChecks");
exports.MAJOR = 'MAJOR';
exports.MINOR = 'MINOR';
exports.PATCH = 'PATCH';
var SemanticVersionSection;
(function (SemanticVersionSection) {
    SemanticVersionSection[SemanticVersionSection["MAJOR"] = 0] = "MAJOR";
    SemanticVersionSection[SemanticVersionSection["MINOR"] = 1] = "MINOR";
    SemanticVersionSection[SemanticVersionSection["PATCH"] = 2] = "PATCH";
})(SemanticVersionSection = exports.SemanticVersionSection || (exports.SemanticVersionSection = {}));
var VersionComparisonResult;
(function (VersionComparisonResult) {
    VersionComparisonResult[VersionComparisonResult["LESS_THAN"] = 0] = "LESS_THAN";
    VersionComparisonResult[VersionComparisonResult["EQUAL_TO"] = 1] = "EQUAL_TO";
    VersionComparisonResult[VersionComparisonResult["GREATER_THAN"] = 2] = "GREATER_THAN";
})(VersionComparisonResult = exports.VersionComparisonResult || (exports.VersionComparisonResult = {}));
function ValidateVersion(version) {
    if (TypeChecks_1.IsString(version)) {
        const sections = version.split('.');
        ValidateVersionArray(sections, 'Version validation failed.');
        return;
    }
    else if (TypeChecks_1.IsArray(version)) {
        ValidateVersionArray(version, 'Version validation failed.');
        return;
    }
    throw new SemanticVersionException(`Failed to validate the semantic version. ${version} is not a valid string or array.`);
}
exports.ValidateVersion = ValidateVersion;
function VersionComparison(left, right) {
    const leftSections = VersionToArray(left);
    ValidateVersionArray(leftSections);
    const rightSections = VersionToArray(right);
    ValidateVersionArray(rightSections);
    for (let i = 0; i < 3; i++) {
        const leftValue = Number(leftSections[i]);
        const rightValue = Number(rightSections[i]);
        if (leftValue < rightValue) {
            return VersionComparisonResult.LESS_THAN;
        }
        else if (leftValue > rightValue) {
            return VersionComparisonResult.GREATER_THAN;
        }
    }
    return VersionComparisonResult.EQUAL_TO;
}
exports.VersionComparison = VersionComparison;
function IncrementMajor(version) {
    if (TypeChecks_1.IsString(version)) {
        return IncrementVersionString(version, 0);
    }
    else if (TypeChecks_1.IsArray(version)) {
        return IncrementVersionArray(version, 0);
    }
    throw new SemanticVersionException(`Failed to increment the ${exports.MAJOR} version. ${version} is not a valid string or array.`);
}
exports.IncrementMajor = IncrementMajor;
function IncrementMinor(version) {
    if (TypeChecks_1.IsString(version)) {
        return IncrementVersionString(version, 1);
    }
    else if (TypeChecks_1.IsArray(version)) {
        return IncrementVersionArray(version, 1);
    }
    throw new SemanticVersionException(`Failed to increment the ${exports.MINOR} version. ${version} is not a valid string or array.`);
}
exports.IncrementMinor = IncrementMinor;
function IncrementPatch(version) {
    if (TypeChecks_1.IsString(version)) {
        return IncrementVersionString(version, 2);
    }
    else if (TypeChecks_1.IsArray(version)) {
        return IncrementVersionArray(version, 2);
    }
    throw new SemanticVersionException(`Failed to increment the ${exports.PATCH} version.
    ${version} is not a valid string or array.`);
}
exports.IncrementPatch = IncrementPatch;
class SemanticVersionException extends Error {
    constructor(message) {
        super(message);
    }
}
exports.SemanticVersionException = SemanticVersionException;
function IncrementVersionArray(version, section) {
    if (!TypeChecks_1.CanBeAPositiveInteger(version[section])) {
        const sectionName = SectionIndexToString(section);
        throw new SemanticVersionException(`Failed to increment the ${sectionName} version.
      The supplied semantic version does not have a value in this section.`);
    }
    const incrementedSection = String(Number(version[section]) + 1);
    const sectionArray = version.map(part => part);
    sectionArray.splice(section, 1, incrementedSection);
    ValidateVersionArray(sectionArray, `Failed to increment the ${SectionIndexToString(section)} version.`);
    return sectionArray.join('.');
}
function IncrementVersionString(version, section) {
    const sections = version.split('.');
    if (!TypeChecks_1.CanBeAPositiveInteger(sections[section])) {
        const sectionName = SectionIndexToString(section);
        throw new SemanticVersionException(`Failed to increment the ${sectionName} version.
      The supplied semantic version does not have a value in this section.`);
    }
    const incrementedSection = String(Number(sections[section]) + 1);
    const sectionArray = sections.map(part => part);
    sectionArray.splice(section, 1, incrementedSection);
    ValidateVersionArray(sectionArray, `Failed to increment the ${SectionIndexToString(section)} version.`);
    return sectionArray.join('.');
}
function SectionIndexToString(section) {
    switch (section) {
        case 0:
            return exports.MAJOR;
        case 1:
            return exports.MINOR;
        case 2:
            return exports.PATCH;
    }
}
function VersionToArray(version) {
    if (TypeChecks_1.IsString(version)) {
        return version.split('.');
    }
    else if (TypeChecks_1.IsArray(version)) {
        return version;
    }
    throw new SemanticVersionException(`Failed to convert the version into an array. The supplied version: ${version} is not a valid string or array.`);
}
function ValidateVersionArray(version, prefix) {
    const versionLength = version.length;
    if (versionLength !== 3) {
        const errorMessagePrefix = prefix ? `${prefix}\n` : 'Invalid semantic version detected.\n';
        const errorMessageDescription = `The following version ${versionLength > 3 ? 'has too many sections.' : 'does not have enough sections.'}.`;
        throw new SemanticVersionException(`${errorMessagePrefix}${errorMessageDescription}`);
    }
    for (let i = 0; i < version.length; i++) {
        if (!TypeChecks_1.CanBeAPositiveInteger(version[i])) {
            const sectionName = SectionIndexToString(i);
            const errorMessagePrefix = prefix ? `${prefix}\n` : 'Invalid semantic version detected.\n';
            const errorMessageDescription = `The ${sectionName} version contains an invalid value: ${version[i]}`;
            throw new SemanticVersionException(`${errorMessagePrefix}${errorMessageDescription}`);
        }
    }
}
