"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
function __export(m) {
    for (var p in m) if (!exports.hasOwnProperty(p)) exports[p] = m[p];
}
Object.defineProperty(exports, "__esModule", { value: true });
const Nano = require("nano");
const BulkTool_1 = require("./BulkTool/BulkTool");
const DesignDocSync_1 = require("./Design/DesignDocSync");
const to_1 = require("./to");
const AggregateView_1 = require("./View/AggregateView");
const ViewEngine_1 = require("./View/ViewEngine");
const COUCH_USER = process.env.COUCH_USER || 'admin';
const COUCH_PASSWORD = process.env.COUCH_PASSWORD || 'password';
const COUCH_PORT = process.env.COUCH_PORT || '5984';
const COUCH_HOST = process.env.COUCH_HOST || 'couchdb';
const COUCH_PROTO = process.env.COUCH_PROTO || 'http';
const CONNECTION_STRING = `${COUCH_PROTO}://${COUCH_USER}:${COUCH_PASSWORD}@${COUCH_HOST}:${COUCH_PORT}`;
const nano = Nano;
class Instance {
    static get Instance() {
        return this._instance;
    }
}
Instance._instance = nano.default ? nano.default(CONNECTION_STRING) : nano(CONNECTION_STRING);
exports.Instance = Instance;
class Couch {
    constructor(databaseName, connection) {
        this.connection = connection ? connection : Instance.Instance;
        this.dbName = databaseName;
        this.initErrorCount = 0;
        this.ready = this.init();
    }
    init() {
        return __awaiter(this, void 0, void 0, function* () {
            const [err, dbList] = yield to_1.default(this.connection.db.list());
            if (Array.isArray(dbList)) {
                if (!DatabaseExists(this.dbName, dbList)) {
                    yield to_1.default(this.connection.db.create(this.dbName));
                }
                this.db = this.connection.use(this.dbName);
                this.bulkTool = new BulkTool_1.BulkToolProvider(this.db);
                return;
            }
            if (this.initErrorCount < 3) {
                console.log(`Error with ${this.dbName}\n`);
                this.initErrorCount++;
                yield this.init();
                return;
            }
            const errorMessage = `An error was encountered while trying to list the available databases. ${err ? err : ''}`;
            throw new Error(errorMessage);
        });
    }
    find(selector) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.ready;
            return yield this.db.find(selector);
        });
    }
    get(id) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.ready;
            return yield this.db.get(id);
        });
    }
    insert(data, id) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.ready;
            if (id) {
                return yield this.db.insert(data, id);
            }
            return yield this.db.insert(data);
        });
    }
    syncDesignDocs(manifest) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.ready;
            const designDocSync = new DesignDocSync_1.DesignDocSync(this, manifest);
            yield designDocSync.sync();
        });
    }
    viewEngine(designDocument, aggregateViews) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.ready;
            const views = aggregateViews.map(view => AggregateView_1.AggregateViewFactory.Create(view));
            const engine = new ViewEngine_1.AggregateViewEngine(this.db, designDocument, views);
            return engine;
        });
    }
    view(designDocument, view, options) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.ready;
            const { rows } = options
                ? yield this.db.view(designDocument, view, options)
                : yield this.db.view(designDocument, view);
            return rows;
        });
    }
    update(data, id) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.ready;
            return yield new Promise(done => {
                this.db.get(id, (error, existing) => {
                    if (error) {
                        done(false);
                        return;
                    }
                    data._rev = existing._rev;
                    this.db.insert(data, id, () => {
                        done(data);
                    });
                });
            });
        });
    }
    patch(data, id) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.ready;
            return yield new Promise(done => {
                this.db.get(id, (error, existing) => {
                    if (error) {
                        done(false);
                        return;
                    }
                    data._rev = existing._rev;
                    const newObj = Object.assign({}, existing, data);
                    this.db.insert(newObj, id, () => {
                        done(newObj);
                    });
                });
            });
        });
    }
    bulkDelete(docs) {
        return this.bulkTool.bulkDelete(docs);
    }
    bulkUpdate(docs) {
        return this.bulkTool.bulkUpdate(docs);
    }
    delete(id) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.ready;
            const [err, data] = yield to_1.default(this.db.get(id));
            if (err)
                return false;
            const [err2] = yield to_1.default(this.db.destroy(id, data._rev));
            return !Boolean(err2);
        });
    }
    get instance() {
        return this.instance;
    }
}
exports.Couch = Couch;
function DatabaseExists(databaseName, availableDatabases) {
    return availableDatabases.includes(databaseName);
}
const nanoInstance = Instance.Instance;
exports.nanoInstance = nanoInstance;
__export(require("./BulkTool"));
__export(require("./CouchProvider"));
__export(require("./Design"));
__export(require("./Interfaces"));
__export(require("./Mock"));
__export(require("./View"));
